/* Copyright (C) 2008 DDengine.org

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; version 2 of the License.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#ifndef _double_table_handler_h
#define _double_table_handler_h

#ifdef USE_PRAGMA_INTERFACE
#pragma interface			/* gcc class implementation */
#endif

#include "revision_handler.h"

class ha_revision;
class single_table_handler;

/** @brief
  Double table revision handler: versioned data is stored in another table.

  Three fields are added to the table:
  - revision_id: incremented on a per row basis when the row is updated.
  - revision_timestamp: set to now() when a row is inserted or updated.
  - revision_deleted: indicates that the row is deleted.

  The revision_id field is added as well to the possible keys of the table.
*/
class double_table_handler : public revision_handler
{

  TABLE *table;                        ///< Opened table
  handler *uhandler;                   ///< Underlying storage engine
  ha_revision *rev_handler;            ///< Revision table handler
  single_table_handler *table_handler; ///< Table handler

  select_mode_t select_mode;
  my_time_t timestamp;
  MEM_ROOT *mem_root;
  bool opened;                       ///< Underlying handler flags

public:
  double_table_handler(TABLE_SHARE *table_arg, MEM_ROOT *mem_root);
  ~double_table_handler();

  /** Used for opening tables. The name will be the name of the file.
  */
  virtual int open(TABLE *table, const char *name,
                   int mode, uint test_if_locked);

  /** Closes a table.
  */
  virtual int close();

  /** Reset state of file to after 'open'
      This function is called after every statement for all tables used
      by that statement.
  */
  virtual int reset();

  /** Creates the table.
  */
  virtual int create(const char *name, TABLE *table_arg,
                     HA_CREATE_INFO *create_info);

  /** Inserts a row.
  */
  virtual int write_row(uchar * buf);

  /** Updates the contents of an existing row.
  */
  virtual int update_row(const uchar * old_data, uchar * new_data);

  /** Deletes a row.
  */
  virtual int delete_row(const uchar * buf);

  /** Handler-specific error messages.
  */
  virtual bool get_error_message(int error, String *buf);

  /* Returns the underlying handler.
  */
  virtual handler *get_uhandler()
  {
    return uhandler;
  }

  /** Index read method.
  */
  virtual int index_read(uchar * buf, const uchar * key,
                         uint key_len, enum ha_rkey_function find_flag);

  /** Used to read forward through the index.
  */
  virtual int index_next(uchar * buf);

  /** Table scan method.
  */
  virtual int rnd_next(uchar *buf);

  /** Called in test_quick_select to determine if indexes should be used.
  */
  virtual double read_time(uint index, uint ranges, ha_rows rows);

  /** Called in test_quick_select to determine if indexes should be used.
  */
  virtual double scan_time();

  virtual uint lock_count(void) const;

  virtual int external_lock(THD *thd, int lock_type);

  virtual THR_LOCK_DATA **store_lock(THD *thd, THR_LOCK_DATA **to,
                                     enum thr_lock_type lock_type);

  /** Sets the select mode for subsequent queries.
  */
  virtual void set_select_mode(select_mode_t);

  /** Sets the timestamp to be used when data state at a specific time
      is to be returned.
  */
  virtual void set_timestamp(const MYSQL_TIME &);

  /** Delete tables.
  */
  virtual void delete_table(const char *, TABLE_SHARE *);

private:
  void create(TABLE_SHARE *table_arg);
  int open_table();

  double_table_handler(const double_table_handler &);
  double_table_handler &operator=(const double_table_handler &);
};

#endif
